#include "ros/ros.h"
#include "std_msgs/MultiArrayLayout.h"
#include "std_msgs/MultiArrayDimension.h"
#include "std_msgs/Int8.h"
#include "std_msgs/Int16.h"
#include <stdio.h>

using namespace std;

#define LEFT_REF_VEL 0
#define LEFT_MAX_VEL 255
#define LEFT_MIN_VEL 0

#define RIGHT_REF_VEL 0
#define RIGHT_MAX_VEL 255
#define RIGHT_MIN_VEL 0

#define LED_REF 0
#define LED_MIN 0
#define LED_MAX 1

#define VELOCITY_INCREMENT 20

// Velocity class that holds different component velocities
class vel_struct{
public:
  
  //initializing velocities constructor
  vel_struct();		
  
  //Setting velocities function
  void setVel( int, int, int, int, int );
  
  int right_f; 		//right forward velocity
  int left_f; 		//left forward velocity
  int right_r; 		//right reverse velocity
  int left_r;  		//right reverse velocity
  int led;   		//led value
};

// Initializing velocities
vel_struct::vel_struct(){
  right_f = 0;
  left_f = 0;
  right_r = 0;
  left_r = 0;
  led = 0;
}

// Setting velocities
void vel_struct :: setVel( int currRight_f, int currLeft_f, int currRight_r, int currLeft_r, int currLed )	{
  right_f = currRight_f;
  left_f = currLeft_f;
  right_r = currRight_r;
  left_r = currLeft_r;
  led = currLed;
}

// Variable declarations
vel_struct curr_vel;
int prev_mode = 0;
std_msgs::Int16 vel;

// Subscribers and publisher declarations
ros::Subscriber direction_sub;
ros::Publisher velocity_pub;

// Fuction declarations
void publish_vel( int );
void controlCallback(const std_msgs::Int8::ConstPtr&);
void maxminVelocityCheck();

//Main function
int main(int argc, char **argv)	{
  ros::init(argc, argv, "weg_velocity_handler");
  ros::NodeHandle n;

  direction_sub = n.subscribe("weg_dir", 10, controlCallback);
  velocity_pub = n.advertise<std_msgs::Int16>("weg_vel", 10);

  ros::Rate loop_rate(10);
  while (ros::ok()) {
    ros::spinOnce();
    loop_rate.sleep();
  }
  return 0;
}

// Publish velocity values
void publish_vel( int curr_mode )	{ 
  switch(curr_mode)	{
    case 0:
      vel.data = 0;
      break;
    case 1:
    case -1:
      vel.data = 1000+curr_vel.left_f;
      break;
    case 2:
    case -2:
      vel.data = 2000+curr_vel.right_f;
      break;
    case 3:
    case -3:
      vel.data = 3000+curr_vel.left_r;
      break;      
    case 4:
    case -4:
      vel.data = 4000+curr_vel.right_r;
      break; 
    case 9:
    case -9:
      vel.data = 9000+curr_vel.led;
      break; 
  }
  
  cout<<vel<<endl;
  velocity_pub.publish( vel );
}

// Control callback function
void controlCallback(const std_msgs::Int8::ConstPtr& mode)
{
  int curr_mode = mode->data;
  
  // Setting velocity value
  switch(curr_mode)	{
    case 0:
      curr_vel.setVel( RIGHT_REF_VEL, LEFT_REF_VEL, RIGHT_REF_VEL, LEFT_REF_VEL, LED_REF);
      break;
    case 1:
      if (prev_mode == curr_mode && curr_vel.left_f < LEFT_MAX_VEL) curr_vel.left_f+=VELOCITY_INCREMENT;	
      break;
    case -1:
      if (prev_mode == curr_mode && curr_vel.left_f > LEFT_MIN_VEL)	curr_vel.left_f-=VELOCITY_INCREMENT;
      break;
    case 2:
      if (prev_mode == curr_mode && curr_vel.right_f < RIGHT_MAX_VEL)	curr_vel.right_f+=VELOCITY_INCREMENT;
      break;
    case -2:
      if (prev_mode == curr_mode && curr_vel.right_f > RIGHT_MIN_VEL)	curr_vel.right_f-=VELOCITY_INCREMENT;
      break;
    case 3:
      if (prev_mode == curr_mode && curr_vel.left_r < LEFT_MAX_VEL)	curr_vel.left_r+=VELOCITY_INCREMENT;
      break;
    case -3:
      if (prev_mode == curr_mode && curr_vel.left_r > LEFT_MIN_VEL)	curr_vel.left_r-=VELOCITY_INCREMENT;
      break;      
    case 4:
      if (prev_mode == curr_mode && curr_vel.right_r < RIGHT_MAX_VEL)	curr_vel.right_r+=VELOCITY_INCREMENT;
      break;
    case -4:
      if (prev_mode == curr_mode && curr_vel.right_r > RIGHT_MIN_VEL)	curr_vel.right_r-=VELOCITY_INCREMENT;
      break;      
    case 9:
      if (prev_mode == curr_mode && curr_vel.led < LED_MAX)	curr_vel.led++;
      break;
    case -9:
      if (prev_mode == curr_mode && curr_vel.led > LED_MIN)	curr_vel.led--;
      break;      
  }
  prev_mode = curr_mode;
  
  // Checks velocity limits
  maxminVelocityCheck();
  
  // Publish velocities
  publish_vel(curr_mode); 
}

// Check for velocity limits and resets them if over or under
void maxminVelocityCheck() {
  if (curr_vel.left_f > LEFT_MAX_VEL) curr_vel.left_f = LEFT_MAX_VEL;
  if (curr_vel.left_f < LEFT_MIN_VEL) curr_vel.left_f = LEFT_MIN_VEL;

  if (curr_vel.right_f > RIGHT_MAX_VEL) curr_vel.right_f = RIGHT_MAX_VEL;
  if (curr_vel.right_f < RIGHT_MIN_VEL) curr_vel.right_f = RIGHT_MIN_VEL;

  if (curr_vel.left_r > LEFT_MAX_VEL) curr_vel.left_r = LEFT_MAX_VEL;
  if (curr_vel.left_r < LEFT_MIN_VEL) curr_vel.left_r = LEFT_MIN_VEL;

  if (curr_vel.right_r > RIGHT_MAX_VEL) curr_vel.right_r = RIGHT_MAX_VEL;
  if (curr_vel.right_r < RIGHT_MIN_VEL) curr_vel.right_r = RIGHT_MIN_VEL;  
}
